﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Security.Claims;
using System.Text.Json;
using System.Threading;
using System.Threading.Tasks;
using InternshipSystem.Api.Options;
using InternshipSystem.Api.Security;
using InternshipSystem.Core;
using InternshipSystem.Repository;
using Microsoft.AspNetCore.Authorization;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.Options;
using Serilog;

namespace InternshipSystem.Api.Controllers
{
    [Route("access")]
    [ApiController]
    public class AccessController : ControllerBase
    {
        private readonly InternshipDbContext _context;
        private readonly JwtTokenService _tokenService;
        private readonly GutCasClient _loginClient;
        private readonly SecurityOptions _securityOptions;

        public AccessController(
            IOptions<SecurityOptions> options,
            InternshipDbContext context,
            JwtTokenService tokenService,
            GutCasClient loginClient)
        {
            _context = context;
            _tokenService = tokenService;
            _loginClient = loginClient;
            _securityOptions = options.Value;
        }
        

        [HttpGet("login")]
        public async Task<ActionResult> Authenticate(string code, CancellationToken cancellationToken)
        {
            var token =  await _loginClient.GetCasTokenAsync(code, cancellationToken);

            var casData = await _loginClient.GetProfileAsync(token, cancellationToken);

            if (!long.TryParse(casData.PersonNumber, out var id))
            {
                return BadRequest();
            }

            var student = await _context.Students.FirstOrDefaultAsync(s => s.Id == id);

            if (student == null)
            {
                student = CreateStudentWithCasData(casData);
                await _context.Students.AddAsync(student, cancellationToken);
                await _context.SaveChangesAsync(cancellationToken);
            }
            
            var identity = new ClaimsIdentity(new[]
            {
                new Claim(ClaimTypes.Name, student.FirstName),
                new Claim(ClaimTypes.Surname, student.LastName),
                new Claim(InternshipClaims.PersonNumber, student.Id.ToString()) 
            });

            return Ok(_tokenService.generateToken(identity));
        }

        [HttpGet("/dev/login")]
        public async Task<ActionResult> Authenticate(CancellationToken cancellationToken)
        {
            var identity = new ClaimsIdentity(new[]
            {
                new Claim(ClaimTypes.Name, "firstname"),
                new Claim(ClaimTypes.Surname, "lastname"),
                new Claim(InternshipClaims.PersonNumber, "1") 
            });
            
            return Ok(_tokenService.generateToken(identity));
        }

        [HttpGet("loginEdition")]
        [Authorize]
        public async Task<ActionResult> LoginIntoEdition(Guid editionId, [FromServices] User user, CancellationToken token)
        {
            var edition = await _context.Editions.FindAsync(editionId);

            var hasInternship = await _context.Entry(edition)
                .Collection(e => e.Internships)
                .Query()
                .AnyAsync(i => i.Student.Id == user.PersonNumber, token);

            if (!hasInternship)
            {
                return Unauthorized("Student isn't registered for this edition");
            }

            var newIdentity = User.Identities.First();
            
            newIdentity.AddClaim(new Claim(InternshipClaims.Edition, editionId.ToString()));

            return Ok(_tokenService.generateToken(newIdentity));
        }

        private Student CreateStudentWithCasData(CasUserData casData)
        {
            var id = long.Parse(casData.PersonNumber);
            var firstName = casData.FirstName;
            var lastName = casData.LastName;
            var email = casData.Mail.First(s => s.EndsWith("@student.pg.edu.pl"));
            var albumNumber = int.Parse(casData.AlbumNumber);

            return Student.CreateStudent(id, firstName, lastName, email, albumNumber);
        }
    }

    public class GutCasClient
    {
        private readonly HttpClient _client;
        private readonly SecurityOptions _securityOptions;

        public GutCasClient(HttpClient client, IOptions<SecurityOptions> options)
        {
            _securityOptions = options.Value;

            client.BaseAddress = _securityOptions.BaseUrl;
            _client = client;
        }

        public async Task<string> GetCasTokenAsync(string code, CancellationToken cancellationToken)
        {
            var request = new HttpRequestMessage
            {
                Method = HttpMethod.Post,
                Content = new FormUrlEncodedContent(new Dictionary<string, string>
                {
                    { "grant_type", "authorization_code" },
                    { "client_id", _securityOptions.ClientId },
                    { "client_secret", _securityOptions.Secret },
                    { "redirect_uri", _securityOptions.RedirectUrl.ToString() },
                    { "code", code }
                }),
                RequestUri = _securityOptions.TokenPath
            };

            var response = await _client.SendAsync(request, cancellationToken);
            await using var stream = await response.Content.ReadAsStreamAsync();

            var value = await JsonSerializer.DeserializeAsync<Dictionary<string, object>>(stream);

            return value["access_token"].ToString();
        }


        public async Task<CasUserData> GetProfileAsync(string token, CancellationToken cancellationToken)
        {
            var request = new HttpRequestMessage
            {
                Method = HttpMethod.Get,
                Content = new StringContent(string.Empty),
                RequestUri = _securityOptions.ProfilePath
            };
            
            request.Headers.Authorization = AuthenticationHeaderValue.Parse($"Bearer {token}");
            request.Content.Headers.ContentType = MediaTypeHeaderValue.Parse("application/json");

            var response = await _client.SendAsync(request, cancellationToken);
            await using var stream = await response.Content.ReadAsStreamAsync();
            
            var result = await JsonSerializer.DeserializeAsync<CasUserProfile>(
                stream, 
                new JsonSerializerOptions
                {
                    PropertyNameCaseInsensitive = true
                });

            return result.Attributes;
        }
    }
}